@ mode4dc.asm
@ mic, 2005 | micol972@gmail.com

@ Uses two 256-color text BGs and a direct color bitmap in mode 4

.text
.global main
.global forever


main:

@ Write some junk to DISPCNT (just for fun..)
ldr r0,=0x11340
mov r1,#0x4000000
str r0,[r1]


@ Turn on both screens, main screen on top
ldr r0,=0x830F
ldr r1,=0x4000304
str r0,[r1]


@ Map VRAM bank A to main BG, bank C to sub BG
mov r0,#0x81
ldr r1,=0x4000240
strb r0,[r1],#2
add r0,r0,#3
strb r0,[r1]


@ Set up rotation/scaling parameters for BG3
ldr r0,=0x4000030
mov r2,#0
mov r1,#0x100
strh r1,[r0],#2
strh r2,[r0],#2
strh r2,[r0],#2
strh r1,[r0],#2
@mov r2,#0
ldr r2,=0xFFFFC000
str r2,[r0],#4
ldr r2,=0xFFFFE000
str r2,[r0],#4


@ Set up palette
mov r0,#0
mov r1,#0x5000000
add r9,r1,#0x400
mov r7,#256
mov r8,#384
set_palette:
	mov r2,r0,lsr#1		@ 0..31
	rsb r3,r2,#31		@ 31..0

	strh r2,[r1]
	strh r2,[r9]
	orr r4,r3,r2,lsl#5
	strh r4,[r1,#128]
	strh r4,[r9,#128]
	mov r4,r2,lsl#10
	orr r4,r4,r3,lsl#5
	strh r4,[r1,r7]
	strh r4,[r9,r7]
	mov r4,r3,lsl#10
	strh r4,[r1,r8]
	strh r4,[r9,r8]
	add r1,r1,#2
	add r9,r9,#2
	add r0,r0,#1
	cmp r0,#64
	bne set_palette
	

@ Store some variables for the plasma at offset 0x100000 in RAM
ldr r0,=0x2100000
mov r8,#0
mov r9,#0
mov r10,#0
mov r11,#0
stmia r0!,{r8-r11}

@ Store BG0 and BG1 scrolling variables
stmia r0,{r8-r11}


@ Create 3 tiles 
mov r0,#0
mov r1,#0x6000000
ldr r2,=0x7070		@ Green 
ldr r3,=0xb0b0		@ Blue
ldr r4,=0x3030		@ Red
create_tiles:
	strh r2,[r1]
	strh r3,[r1,#64]
	strh r4,[r1,#128]
	add r1,r1,#2
	add r0,r0,#1
	cmp r0,#32
	bne create_tiles
	
@ Now fill BG0, BG1 and Sub BG0 with some tiles	
ldr r0,=0x6004000
mov r8,#0x0200000
mov r9,#0x2000
mov r1,#0
fill_bg0_vert:
	mov r2,#0
	fill_bg0_horz:
		eor r3,r1,r2
		mov r3,r3,lsr#1
		and r3,r3,#1
		strh r3,[r0]
		add r3,r3,#1
		strh r3,[r0,r9]
		mov r3,#0
		strh r3,[r0,r8]		@ Clear Sub BG0
		add r0,r0,#2
		add r2,r2,#1
		cmp r2,#32
		bne fill_bg0_horz
	add r1,r1,#1
	cmp r1,#32
	bne fill_bg0_vert


@ Copy font data
ldr r0,=font
ldr r1,=0x6200000
ldr r2,=2688
copy_font:
	ldrh r3,[r0],#2
	strh r3,[r1],#2
	subs r2,r2,#1
	bne copy_font
	

@ Set BG mode 4 and enable BG0 + BG1 + BG3
ldr r0,=0x0010B04
mov r1,#0x4000000
str r0,[r1]	

@ BG0: text bg, 256 colors, 256x256 pixels, screen base 0x4000, priority 2
ldr r0,=0x0882
ldr r1,=0x4000008
strh r0,[r1]

@ BG1: text bg, 256 colors, 256x256 pixels, screen base 0x6000, priority 1
ldr r0,=0x0C81
ldr r1,=0x400000a
strh r0,[r1]

@ BG3: direct color mode, 128x128 pixels, screen base 0x8000, priority 0
ldr r0,=0x284
ldr r1,=0x400000e
strh r0,[r1]


@ Set up BLDCNT and BLDALPHA
ldr r0,=0x4000050
ldr r1,=0xF4F	@ Alpha-blend, all layers
strh r1,[r0],#2
ldr r1,=0x90F	@ 1st target param 15, 2nd target param 9
strh r1,[r0]



@@@@ Sub core setup @@@@

@ Set mode 0 and enable BG0
ldr r0,=0x0010100
ldr r1,=0x4001000
str r0,[r1]	

@ Sub BG0: text bg, 256 colors, 256x256 pixels, screen base 0x4000, priority 0
ldr r0,=0x0880
ldr r1,=0x4001008
strh r0,[r1]

@ Print some text on Sub BG0
bl VSync
ldr r0,=szText
mov r1,#32
mov r2,#88
mov r3,#0x4000
bl DrawText



@@@@ Main loop @@@@


forever:
	bl VSync

	@ Load trig table offsets 3 and 4
	ldr r12,=0x210000C
	ldr r11,[r12],#-4
	ldr r10,[r12],#-4

	mov r0,#0x6000000
	add r0,r0,#0x8000
	ldr r1,=trigtable
	mov r2,#128
	mov r8,#256
	yloop:
		add r10,r10,#1
		add r11,r11,#2
		and r10,r10,#0xFF
		and r11,r11,#0xFF
		@ Load trig table offset 1 and 2
		ldr r6,[r12,#-4]
		ldr r7,[r12]
		mov r3,#128
		xloop:
			add r6,r6,#2
			add r7,r7,#3
			and r6,r6,#0xFF
			and r7,r7,#0xFF
			ldrb r4,[r1,r6]
			ldrb r5,[r1,r7]
			add r4,r4,r5
			ldrb r5,[r1,r10]
			add r4,r4,r5
			ldrb r5,[r1,r11]
			add r4,r4,r5
			and r4,r4,#0xFF
			mov r4,r4,lsr#3
			mov r5,r4,lsr#1
			orr r4,r4,#0x8000
			orr r4,r4,r5,lsl#5
			strh r4,[r0]
			subs r3,r3,#1
			add r0,r0,#2
			bne xloop
		@add r0,r0,#256
		subs r2,r2,#1
		bne yloop

	@ Update plasma variables
	ldr r0,=0x2100000
	ldr r1,[r0]
	sub r1,r1,#3
	str r1,[r0],#4
	

	ldr r1,[r0]
	add r1,r1,#2
	str r1,[r0],#4
	

	ldr r1,[r0]
	sub r1,r1,#3
	str r1,[r0],#4
	

	ldr r1,[r0]
	add r1,r1,#1
	str r1,[r0],#4
	
	
	@ Update H/VOFS for BG0
	ldr r2,=0x4000010
	ldr r1,[r0]
	add r1,r1,#1
	str r1,[r0],#4
	strh r1,[r2],#2
	ldr r1,[r0]
	add r1,r1,#1
	str r1,[r0],#4
	strh r1,[r2],#2

	@ Update H/VOFS for BG1
	ldr r2,=0x4000014
	ldr r1,[r0]
	sub r1,r1,#1
	str r1,[r0],#4
	strh r1,[r2],#2
	ldr r1,[r0]
	sub r1,r1,#1
	str r1,[r0],#4
	strh r1,[r2],#2
	
	b forever



VSync:
  stmfd sp!,{r0-r1}
  mov r0,#0x4000000
  add r0,r0,#4		@ DISPCNT+4 = DISPSTAT

  vs_loop1:
    ldrh r1,[r0]
    ands r1,r1,#1
  bne vs_loop1

  vs_loop2:
    ldrh r1,[r0]
    ands r1,r1,#1
  beq vs_loop2

  ldmfd sp!,{r0-r1}
  mov pc,lr


DrawText:
@ r0: lpszText
@ r1: x
@ r2: y
@ r3: screen base
	stmfd sp!,{r4-r10}

	mov r9,r2,lsr#3
	mov r9,r9,lsl#6
	mov r7,r1,lsr#3
	add r9,r9,r3
	add r9,r9,r7,lsl#1
	add r9,r9,#0x6000000
	add r9,r9,#0x0200000
	
dt_cloop:	
	ldrb r4,[r0],#1
	cmp r4,#0
	beq dt_null
	cmp r4,#32
	moveq r4,#0
	subne r4,r4,#37

	strh r4,[r9],#2	
	
	b dt_cloop
dt_null:
	ldmfd sp!,{r4-r10}
	mov pc,lr

.pool
.align



.align 2
.pool

.align 2
font:
	.incbin "font_grad.pat"

.align 2	
trigtable:
.byte 64,63,63,63,63,63,63,63,63,63,63,62,62,62,62,61
.byte 61,61,60,60,60,59,59,58,58,58,57,57,56,56,55,55
.byte 54,53,53,52,52,51,50,50,49,49,48,47,46,46,45,44
.byte 44,43,42,41,41,40,39,38,38,37,36,35,34,34,33,32
.byte 31,31,30,29,28,27,27,26,25,24,24,23,22,21,20,20
.byte 19,18,18,17,16,15,15,14,13,13,12,12,11,10,10,9
.byte 9,8,8,7,7,6,6,5,5,4,4,3,3,3,2,2
.byte 2,2,1,1,1,1,0,0,0,0,0,0,0,0,0,0
.byte 0,0,0,0,0,0,0,0,0,0,1,1,1,1,2,2
.byte 2,2,3,3,3,4,4,5,5,6,6,7,7,8,8,9
.byte 9,10,10,11,12,12,13,13,14,15,16,16,17,18,18,19
.byte 20,20,21,22,23,24,24,25,26,27,27,28,29,30,31,31
.byte 32,33,34,34,35,36,37,38,38,39,40,41,41,42,43,44
.byte 44,45,46,46,47,48,49,49,50,50,51,52,52,53,53,54
.byte 55,55,56,56,57,57,58,58,58,59,59,60,60,60,61,61
.byte 61,62,62,62,62,63,63,63,63,63,63,63,63,63,63,64
 
 .align 2
 szText: .asciz "MODE 4 DIRECT COLOR DEMO" 

.align
.pool
.end
